/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (03/2001)
 */

#include "global.h"
#include "cocoa_input.h"
#include "pixmaps.h"
#include <qaccel.h>
#include <qfiledialog.h>
#include <qfileinfo.h>
#include <qiodevice.h>
#include <qmessagebox.h>
#include <qtextstream.h>

CocoaInput::CocoaInput(CocoaOutput *_engine, CocoaOutput *_cocoaOutput, QWidget *parent, const char *name)
	: QTabWidget(parent, name), engine(_engine), cocoaOutput(_cocoaOutput), autoCompleting(false), autoIndenting(false)
{
	/* Just to make sure that currentFont is initialized */
	currentFont = QFont("Courier");

	/* This tells the list to delete documents when they are removed from the list */
	documents.setAutoDelete(true);

	interactiveDocument = new InteractiveDocument(engine, cocoaOutput, this);

	connect(interactiveDocument, SIGNAL(execute(QString, bool, ButtonState)), this, SLOT(executeCommandSet(QString, bool, ButtonState)));
	connect(interactiveDocument, SIGNAL(cursorPositionChanged(int, int)), this, SIGNAL(cursorPositionChanged(int, int)));

	connect(interactiveDocument, SIGNAL(copyAvailable(bool)), this, SIGNAL(copyAvailable(bool)));
	connect(interactiveDocument, SIGNAL(redoAvailable(bool)), this, SIGNAL(redoAvailable(bool)));
	connect(interactiveDocument, SIGNAL(undoAvailable(bool)), this, SIGNAL(undoAvailable(bool)));

	/*
	 * Workaround for Qt bug on Windows
	 */
	connect(interactiveDocument, SIGNAL(toggleAutoCompletionMode()), this, SIGNAL(toggleAutoCompletionMode()));
	
	modifiedIconSet = createIconSet(modified_xpm);

	addTab(interactiveDocument, "Interactive (&0)");
	connect(this, SIGNAL(selected(const QString &)), this, SLOT(selected(const QString &)));
	interactiveDocument->setFocus();
	interactiveDocument->setFrameStyle(QFrame::Plain);
}

CocoaDocument* CocoaInput::createDocument()
{
	return createDocument(NULL);
}

CocoaDocument* CocoaInput::createDocument(const QString& fileName)
{
	CocoaDocument *doc = new CocoaDocument(fileName, engine, cocoaOutput, this);
	doc->setFrameStyle(QFrame::Plain);
	doc->setEdited(false);
	doc->setFont(currentFont);
	doc->setWordCompletions(wordCompletions);
	doc->setAutoCompleting(autoCompleting);
	doc->setAutoIndenting(autoIndenting);

	connect(doc, SIGNAL(execute(QString, bool, ButtonState)), this, SLOT(executeCommandSet(QString, bool, ButtonState)));
	connect(doc, SIGNAL(cursorPositionChanged(int, int)), this, SIGNAL(cursorPositionChanged(int, int)));

	connect(doc, SIGNAL(copyAvailable(bool)), this, SIGNAL(copyAvailable(bool)));
	connect(doc, SIGNAL(redoAvailable(bool)), this, SIGNAL(redoAvailable(bool)));
	connect(doc, SIGNAL(undoAvailable(bool)), this, SIGNAL(undoAvailable(bool)));
	connect(doc, SIGNAL(saveAvailable(CompletingMultiLineEdit*, bool)), this, SLOT(emitSaveAvailable(CompletingMultiLineEdit*, bool)));

	/*
	 * Workaround for Qt bug on Windows
	 */
	connect(doc, SIGNAL(toggleAutoCompletionMode()), this, SIGNAL(toggleAutoCompletionMode()));

	documents.append(doc);
	addTab(doc, createDisplayName(doc->getFileName(), documents.count()));
	showPage(doc);
	if (fileName.isEmpty())
		emit setStatus("New document created", 10000);
	return doc;
}

void CocoaInput::updateDisplayNames()
{
	QListIterator<CocoaDocument> it(documents);
	for (; it.current(); ++it)
	{
		CocoaDocument *doc = it.current();
		int index = getDocumentIndex(doc);
		QIconSet iconSet = doc->canSave() ? modifiedIconSet : QIconSet();
		changeTab(it.current(), iconSet, createDisplayName(it.current()->getFileName(), index, index < 10));
	}
}

void CocoaInput::removeDocument(CocoaDocument *document)
{
	removePage(document);

	/*
	 * This not only removes the document from the list but also destroys it
	 * since we called setAutoDelete( true ) in the constructor.
	 */
	documents.remove(document);

	updateDisplayNames();
}

bool CocoaInput::closeAllDocuments()
{
	CocoaDocument *doc;
	QListIterator<CocoaDocument> it(documents);
	QList<CocoaDocument> documentsToClose;
	bool success = true;
	for (; doc = it.current(); ++it)
	{
		success = closeDocument(doc, false);
		if (success)
			documentsToClose.append(doc);
		else
			break;
	}
	QListIterator<CocoaDocument> it2(documentsToClose);
	for (; doc = it2.current(); ++it2)
		removeDocument(doc);

	return success;
}

bool CocoaInput::closeCurrentDocument()
{
	if (currentEditor() == interactiveDocument)
	{
		emit setStatus("The interactive document cannot be closed", 10000);
		return false;
	}

	return closeDocument(currentEditor());
}

bool CocoaInput::closeDocument(CocoaDocument *doc, bool removeFromTab)
{
	if (doc == interactiveDocument)
	{
		emit setStatus("The interactive document cannot be closed", 10000);
		return false;
	}

	if (doc->edited())
	{
		QString displayName = createDisplayName(doc->getFileName(), getDocumentIndex(doc), false);
		int response = QMessageBox::information(this, QString(PROGRAM_NAME) + " " + XCOCOA_VERSION,
				QString("The document ") + displayName + " has been modified.\n"
				"Would you like to save it?",
				"Yes", "No", "Cancel", 0, 1);
		switch(response)
		{
			case 0:
			{
				bool success = saveDocument(doc);
				if (success && removeFromTab)
				{
					removeDocument(doc);
					emit setStatus(displayName + " has been closed", 10000);
				}
				return success;
				break;
			}
			case 1:
				if (removeFromTab)
					removeDocument(doc);
				emit setStatus(displayName + " has been closed", 10000);
				return true;
				break;
			case 2:
				emit setStatus(displayName + " has not been closed", 10000);
				return false;
				break;
		}
	} else if (removeFromTab)
		removeDocument(doc);

	emit setStatus("No unsaved changes. Document closed.", 10000);
	return true;
}

void CocoaInput::selected(const QString& tabName)
{
	QWidget *p = currentPage();
	if (p == interactiveDocument)
		emit documentSelected("Interactive Document");
	else
		emit documentSelected(((CocoaDocument *)p)->getFileName());

	CocoaDocument *editor = currentEditor();
	emit redoAvailable(editor->canRedo());
	emit undoAvailable(editor->canUndo());
	emit copyAvailable(editor->canCopy());

	int line, column;
	editor->getCursorPosition(&line, &column);
	emit cursorPositionChanged(line, column);

	if (editor != interactiveDocument)
	{
		emitSaveAvailable(editor, editor->canSave());
		emit closeAvailable(true);
		interactiveDocument->activeStateChanged(false);
	} else
	{
		emitSaveAvailable(editor, false);
		emit closeAvailable(false);
		interactiveDocument->activeStateChanged(true);
	}
}

void CocoaInput::setFont(const QFont& f)
{
	currentFont = f;
	interactiveDocument->setFont(f);

	QListIterator<CocoaDocument> it(documents);
	for (; it.current(); ++it)
		it.current()->setFont(f);
}

void CocoaInput::copy()
{
	CocoaDocument *doc = currentEditor();
	if (doc->canCopy())
		doc->copy();
}

void CocoaInput::cut()
{
	CocoaDocument *doc = currentEditor();
	if (doc->canCopy())
		doc->cut();
}

void CocoaInput::paste()
{
	currentEditor()->paste();
}

void CocoaInput::redo()
{
	CocoaDocument *doc = currentEditor();
	if (doc->canRedo())
		doc->redo();
}

void CocoaInput::undo()
{
	CocoaDocument *doc = currentEditor();
	if (doc->canUndo())
		doc->undo();
}

bool CocoaInput::load()
{
	QString fileName = QFileDialog::getOpenFileName(QString::null, QString::null, this);
	if (fileName.isEmpty())
	{
		emit setStatus("Load aborted", 10000);
		return false;
	}

	return load(fileName);
}

bool CocoaInput::load(QString fileName)
{
	QFile file(fileName);
	if (!file.open( IO_ReadOnly ))
	{
		emit setStatus(QString("Can't open file %1").arg(fileName), 10000);
		return false;
	}

	CocoaDocument *doc = createDocument(fileName);
	doc->setAutoUpdate(false);
	doc->clear();

	QTextStream in(&file);
	while (!in.atEnd())
		doc->append(in.readLine());
	file.close();

	doc->setEdited(false);
	doc->setAutoUpdate(true);
	doc->repaint();
	emit setStatus(QString("Loaded document %1").arg(fileName), 10000);
	return true;
}

bool CocoaInput::save()
{
	CocoaDocument *doc = currentEditor();
	return saveDocument(doc);
}

bool CocoaInput::saveDocument(CocoaDocument *doc)
{
	if (!doc->canSave())
		return false;

	QString fileName = doc->getFileName();

	if (fileName.isEmpty())
		return saveDocumentAs(doc);

	if (!saveAs(doc, fileName))
	{
		emit setStatus(QString("Could not write to %1").arg(fileName), 10000);
		if (QMessageBox::warning(this,
				QString(PROGRAM_NAME) + " " + XCOCOA_VERSION,
				"Could not save document",
				"Save as", "Cancel", 0, 0) == 0)
			return saveDocumentAs(doc);

		return false;
	} else
	{
		emit setStatus(QString("File %1 saved").arg(fileName), 10000);
		return true;
	}
}

bool CocoaInput::saveAs()
{
	CocoaDocument *doc = currentEditor();
	return saveDocumentAs(doc);
}

bool CocoaInput::saveDocumentAs(CocoaDocument *doc)
{
	bool aborted = false;

	while (!aborted)
	{
		QString displayName;
		if (doc == interactiveDocument)
			displayName = "Interactive Document";
		else
		{
			displayName = QFileInfo(doc->getFileName()).fileName();
			if (displayName.isEmpty())
				displayName = createDisplayName("Untitled", getDocumentIndex(doc), false);
		}

		QString fileName;
		bool useFileName = false;
		do
		{
			fileName = QFileDialog::getSaveFileName(QString::null, QString::null, this, 0, "Save " + displayName);
		
			if (fileName.isEmpty())
				aborted = true;
			
			if (!aborted && QFile(fileName).exists())
			{
				if (QMessageBox::warning(this,
					QString(PROGRAM_NAME) + " " + XCOCOA_VERSION,
					fileName + " already exists.\n"
					"Do you want to overwrite it?",
					"Yes", "No", 0, 1) == 0)
					useFileName = true;
			} else
				useFileName = true;
		} while (!aborted && !useFileName);

		if (!aborted)
		{
			int success = saveAs(doc, fileName);
			if (success)
			{
				doc->setFileName(fileName);
				updateDisplayNames();
				documentSelected(doc->getFileName());
				emit setStatus(QString("File %1 saved").arg(fileName), 10000);
				return true;
			} else
			{
				emit setStatus(QString("Could not write to %1").arg(fileName), 10000);
				if (QMessageBox::warning(this,
					QString(PROGRAM_NAME) + " " + XCOCOA_VERSION,
					"Could not save document",
					"Save as", "Cancel", 0, 0) == 1)
					aborted = true;
			}
		}
	}

	emit setStatus("Saving aborted", 10000);
	return false;
}

void CocoaInput::execute()
{
	CocoaDocument *doc = currentEditor();
	doc->execInput(ControlButton);
}

void CocoaInput::setAutoCompleting(bool b)
{
	autoCompleting = b;
	CocoaDocument *doc;
	interactiveDocument->setAutoCompleting(b);
	for (doc = documents.first(); doc != 0; doc = documents.next())
		doc->setAutoCompleting(b);
}

void CocoaInput::setAutoIndenting(bool b)
{
	autoIndenting = b;
	CocoaDocument *doc;
	interactiveDocument->setAutoIndenting(b);
	for (doc = documents.first(); doc != 0; doc = documents.next())
		doc->setAutoIndenting(b);
}

void CocoaInput::setWordCompletions(WordCompletions *_wordCompletions)
{
	wordCompletions = _wordCompletions;

	interactiveDocument->setWordCompletions(wordCompletions);
	QListIterator<CocoaDocument> it(documents);
	for (; it.current(); ++it)
		it.current()->setWordCompletions(wordCompletions);
}

CocoaDocument* CocoaInput::currentEditor()
{
	QWidget *p = currentPage();
	return (CocoaDocument*) p;
}

const QString CocoaInput::createDisplayName(const QString& fileName, const int& index, bool showShortcut) const
{
	QString displayName;
	if (!fileName.isEmpty())
		displayName = QFileInfo(fileName).fileName();
	else
		displayName = "Untitled";

	if (index >= 10) showShortcut = false;
	QString indexString = (showShortcut ? QString(" (&%1)") : QString(" (%1)")).arg(index);
	return displayName + indexString;
}

int CocoaInput::getDocumentIndex(CocoaDocument *document)
{
	if (document == interactiveDocument)
		return 0;

	QListIterator<CocoaDocument> it(documents);
	int i = 1;
	for (; it.current(); ++it, i++)
		if (it.current() == document) return i;

	/* This should never happen */
	return -1;
}

void CocoaInput::emitSaveAvailable(CompletingMultiLineEdit *edit, bool b)
{
	CocoaDocument *doc = (CocoaDocument*) edit;
	if (doc != interactiveDocument)
	{
		int index = getDocumentIndex(doc);
		if (b)
			changeTab(doc, modifiedIconSet, createDisplayName(doc->getFileName(), index));
		else
			changeTab(doc, QIconSet(), createDisplayName(doc->getFileName(), index));
	}
	emit saveAvailable(b);
}

bool CocoaInput::saveAs(CocoaDocument* document, const QString& fileName)
{
	QFile file(fileName);
	if (!file.open(IO_WriteOnly))
		return false;

	QTextStream out(&file);
	QIODevice *dev = out.device();
	out << document->text();
	if (dev->status() != IO_Ok)
		return false;
	dev->close();
	if (dev->status() != IO_Ok)
		return false;
	document->setEdited(false);
	return true;
}

void CocoaInput::executeCommandSet(QString commands, bool markedText, ButtonState buttonState)
{
	CocoaDocument *doc = currentEditor();

	/* Do nothing when both ControlButton and AltButton are pressed */
	if ((buttonState & ControlButton) && (buttonState & AltButton))
		return;

	/* Jumps to end if and only if ShiftButton is not pressed and we are not executing marked text */
	interactiveDocument->appendPage(commands, (doc == interactiveDocument), (!(buttonState & ShiftButton) && !markedText));

	/* Executes commands if and only if ControlButton is pressed */
	if (buttonState & ControlButton)
		cocoaOutput->execInput(commands);
}
#include "cocoa_input.moc"
