/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (03/2001)
 */

#include "ca_qt.h"
#include "cocoa_output.h"
#include "global.h"

#include <qapplication.h>
#include <qmessagebox.h>
#include <qscrollbar.h>

CocoaOutput::CocoaOutput(QWidget *parent, const char *name)
	: QtMultiLineEdit(parent, name), execution_running(false), blocked(false)
{
	setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);

	currentOutput = "";
	setReadOnly(true);
	doRedraw = false;
	thread = 0;
	time.start();
}

void CocoaOutput::selectWordWrap(QString type)
{
	if (type == WORD_WRAP_NONE)
	{
		setWordWrap(QtMultiLineEdit::FixedPixelWidth);
		setWrapColumnOrWidth(SHRT_MAX);

		/* The next line is a work around for a Qt bug */
		setTableFlags(Tbl_autoHScrollBar);

		setWrapPolicy(Anywhere);
		updateTableSize();
	} else if (type == WORD_WRAP_NORMAL)
	{
		setWordWrap(QtMultiLineEdit::WidgetWidth);
		setWrapPolicy(QtMultiLineEdit::Anywhere);

		updateTableSize();
	} else if (type == WORD_WRAP_FIXED)
	{
		setWrapColumnOrWidth(80);
		setWordWrap(QtMultiLineEdit::FixedColumnWidth);

		/* The next line is a work around for a Qt bug */
		setTableFlags(Tbl_autoHScrollBar);

		setWrapPolicy(Anywhere);

		updateTableSize();
	}
}

void CocoaOutput::selectFont(const QFont& f)
{
	((QtMultiLineEdit*) this)->setFont(f);
}

bool CocoaOutput::isExecuting()
{
	return thread && execution_running;
}

bool CocoaOutput::isBlocked()
{
	return blocked;
}

void CocoaOutput::engineInterrupted(bool interrupted)
{
	if (thread && isExecuting() && interrupted)
	{
		emit currentStatus("Calculation in progress ... interruption pending!");
		mutex.lock();
		ca_interrupted = true;
		mutex.unlock();
	} else if (!interrupted)
	{
		mutex.lock();
		ca_interrupted = false;
		mutex.unlock();
	}
}

void CocoaOutput::addString(const char *s)
{
	currentOutput += s;

	if (doRedraw && (redrawTime.elapsed() > UPDATE_TIME))
	{
		setAutoUpdate(false);
		setCursorPosition(numLines() - 1, 0);
		end();
		int line, col;
		getCursorPosition(&line, &col);
		insertAt(currentOutput, line, col);
		currentOutput = "";

		setAutoUpdate(true);
		flushOutput();
		repaint(false);
		redrawTime.start();
	}
}

void CocoaOutput::flushOutput()
{
	QScrollBar *scrollBar = verticalScrollBar();
	scrollBar->setValue(scrollBar->maxValue());
}

bool CocoaOutput::canCopy() { return hasMarkedText(); }
bool CocoaOutput::canUndo() { return false; }
bool CocoaOutput::canRedo() { return false; }

void CocoaOutput::execInput(QString commands)
{
	emit currentStatus("Calculation in progress ...");

	/*
	 * The first call to this method is the one loading the CoCoA library.
	 * We don't want this call to show in the output window.
	 * Therefore we check if it is the first call by checking if thread is 0.
	 */
	if (thread)
	{
		addString((const char*) commands);
		addString("\n");
		setAutoUpdate(true);
		flushOutput();
		repaint(false);
	}

	doRedraw = true;
	redrawTime.start();

	if (thread)
		delete thread;

	setExecutionRunning(true);
	time.restart();
	thread = new EngineThread(this, commands.latin1());
	thread->start();
}

void CocoaOutput::setThreadExecutionFlag()
{
	execution_running = false;
	qApp->closeAllWindows();
}

void CocoaOutput::setExecutionRunning(bool on)
{
	if (!on)
	{
		setAutoUpdate(false);
		setCursorPosition(numLines() - 1, 0);
		end();
		int line, col;
		getCursorPosition(&line, &col);
		insertAt(currentOutput, line, col);
		currentOutput = "";

		emit currentStatus("Ready [" + durationString(time.elapsed()) + "]");
		doRedraw = false;
		setAutoUpdate(true);
		flushOutput();
		repaint(false);
		engineInterrupted(false);
	}
	execution_running = on;
	emit executing(on);
}

QString CocoaOutput::durationString(int elapsed)
{
	/*
	 * c[0] = ms
	 * c[1] = s
	 * c[2] = m
	 * c[3] = h
	 */
	int c[4];
	QString n[4] = {"ms", "s", "m", "h"};

	c[1] = elapsed / 1000;
	c[0] = elapsed - c[1] * 1000;
	c[2] = c[1] / 60;
	c[1] = c[1] - c[2] * 60;
	c[3] = c[2] / 60;
	c[2] = c[2] - c[3] * 60;

	int z = 3;
	for (int i = 3; i >= 0 && !c[z]; i--, z--);

	if (z < 0)
		z = 0;

	QString s = QString::number(c[z]);
	s += n[z];
	if (z > 0)
	{
		s += QString(" ");
		s += QString::number(c[z - 1]);
		s += n[z - 1];
	}
	return s;
}

/*
 * Workaround for Qt bug: Ctrl-Home and Ctrl-End don't work in read-only mode
 */
void CocoaOutput::keyPressEvent(QKeyEvent *e)
{
	if (e->state() & ControlButton)
	{
		switch (e->key())
		{
			case Key_Home:
				setCursorPosition(0, 0, e->state() & ShiftButton);
				break;
			case Key_End:
				setCursorPosition(numLines() - 1, lineLength(numLines() - 1), e->state() & ShiftButton);
				break;
			default:
				QtMultiLineEdit::keyPressEvent(e);
				break;
		}
	} else
		QtMultiLineEdit::keyPressEvent(e);
}

void CocoaOutput::customEvent(QCustomEvent *e)
{
	switch(e->type())
	{
		case STRINGADD_EVENT_ID:
			addString(((StringAddEvent*) e)->getString());
			break;
		case FLUSH_EVENT_ID:
			flushOutput();
			break;
		case EXECUTIONSTATE_EVENT_ID:
			setExecutionRunning(((ExecutionStateEvent*) e)->isRunning());
			break;
		case QUIT_EVENT_ID:
			setThreadExecutionFlag();
			break;
		case ENGINE_CRASH_EVENT_ID:
			blocked = true;
			QMessageBox::critical(this, QString(PROGRAM_NAME) + " " + XCOCOA_VERSION
				+ " - Error Handler",
				"Something just caused the CoCoA engine to crash.\n"
				"Please save your input files and the program output and send\n"
				"a bug report to the CoCoA team!\n"
				"To avoid loss of data the CoCoA engine has been stopped. To\n"
				"continue your work, you must restart CoCoA!",
				QMessageBox::Ok | QMessageBox::Default, QMessageBox::NoButton);
			break;
		default:
			/*
			 * HACK:
			 * Maybe we need to call QtMultiLineEdit::customEvent(e) here.
			 */
			break;
	}
}

#include "cocoa_output.moc"
