/*
 * Written by Volker Augustin <snowbird@bigfoot.de> (02/2001)
 */

#include "completing_multilineedit.h"

#include <qregexp.h>

CompletingMultiLineEdit::CompletingMultiLineEdit(QWidget *parent, const char *name)
	:QtMultiLineEdit(parent, name)
{
	completing = false;
	auto_completing = false;
	_copyAvailable = _redoAvailable = _undoAvailable = false;
	_saveAvailable = false;

	old_line = old_column = -1;

	connect(this, SIGNAL(copyAvailable(bool)), this, SLOT(setCopyAvailable(bool)));
	connect(this, SIGNAL(redoAvailable(bool)), this, SLOT(setRedoAvailable(bool)));
	connect(this, SIGNAL(undoAvailable(bool)), this, SLOT(setUndoAvailable(bool)));
	connect(this, SIGNAL(textChanged()), this, SLOT(setSaveAvailable()));
}

void CompletingMultiLineEdit::setCopyAvailable(bool b)
{
	_copyAvailable = b;
}

bool CompletingMultiLineEdit::canCopy()
{
	return _copyAvailable;
}

void CompletingMultiLineEdit::setRedoAvailable(bool b)
{
	_redoAvailable = b;
}

bool CompletingMultiLineEdit::canRedo()
{
	return _redoAvailable;
}

void CompletingMultiLineEdit::setUndoAvailable(bool b)
{
	_undoAvailable = b;
}

bool CompletingMultiLineEdit::canUndo()
{
	return _undoAvailable;
}

void CompletingMultiLineEdit::setSaveAvailable()
{
	bool old_saveAvailable = _saveAvailable;
	_saveAvailable = edited();
	if (old_saveAvailable != _saveAvailable)
		emit saveAvailable(this, _saveAvailable);
}

bool CompletingMultiLineEdit::canSave()
{
	return _saveAvailable;
}

void CompletingMultiLineEdit::setEdited(bool b)
{
	((QtMultiLineEdit*) this)->setEdited(b);
	setSaveAvailable();
}

void CompletingMultiLineEdit::keyPressEvent(QKeyEvent *e)
{
	if (wordCompletions && ((e->key() == Key_Down) || (e->key() == Key_Up)) &&
			((e->state() & ControlButton) || completing))
	{
		if (!completing)
		{
			getCursorPosition(&c_line, &c_col);
			QString *text = getString(c_line);
			int position = text->findRev(QRegExp("[^a-zA-Z0-9_.]"), c_col - text->length() - 1);
			c_length = c_col - position - 1;
			c_head = text->mid(position + 1, c_length);
			c_results = wordCompletions->completionsFor(c_head);

			if (c_results.isEmpty() || ((c_results.count() == 1) && (c_results.first() == c_head)))
				completing = false;
			else
			{
				completing = true;

				if (e->key() == Key_Down)
				{
					c_iterator = c_results.begin();
					start_iterator = c_results.end();
					do
						--start_iterator;
					while ((*start_iterator) == c_head);
					shift_used = false;
				} else
				{
					c_iterator = c_results.end();
					--c_iterator;
					start_iterator = c_results.begin();
					do
						++start_iterator;
					while ((*start_iterator) == c_head);
					shift_used = true;
				}
			}
		}

		/* Do not use else here in conjunction with the above */
		/* The above has to be executed first! */
		if (completing)
		{
			if (e->key() == Key_Down)
			{
				if (shift_used)
				{
					for (int i = 0; i < 2; i++)
					{
						if (i == 1)
							start_iterator = c_iterator;
						++c_iterator;
						if (c_iterator == c_results.end())
							c_iterator = c_results.begin();
					}
				}

				if (c_iterator == c_results.end())
					c_iterator = c_results.begin();

				while (((*c_iterator) == c_head) && (*start_iterator) && !(c_iterator == start_iterator))
				{
					++c_iterator;
					if (c_iterator == c_results.end())
						c_iterator = c_results.begin();
				}

				if (hasMarkedText()) del();

				QString word = (*c_iterator).mid(c_length);
				insert(word, true);

				++c_iterator;
				shift_used = false;
			} else
			{
				if (!shift_used)
				{
					for (int i = 0; i < 2; i++)
					{
						if (i == 1)
							start_iterator = c_iterator;
						if (c_iterator == c_results.begin())
							c_iterator = c_results.end();
						--c_iterator;
					}
				}

				while (((*c_iterator) == c_head) && (*start_iterator) && !(c_iterator == start_iterator))
				{
					if (c_iterator == c_results.begin())
						c_iterator = c_results.end();
					--c_iterator;
				}

				if (hasMarkedText()) del();

				QString word = (*c_iterator).mid(c_length);
				insert(word, true);

				if (c_iterator == c_results.begin())
					c_iterator = c_results.end();

				--c_iterator;
				shift_used = true;
			}
		}
	} else
	{
		bool completed_before = false;
		if (completing && !((e->key() & 0xFFF0) == 0x1020)) {
			if (e->key() == Key_Right)
				deselect();
			else if (e->key() == Key_Left)
			{
				deselect();
				insert("()", false);
			} else
				del();
        		completing = false;
			completed_before = true;
		}

		if (!(completed_before && ((e->key() == Key_Right) || (e->key() == Key_Delete))))
			QtMultiLineEdit::keyPressEvent(e);

		/* Don't complete when just a modifier is pressed */
		if (wordCompletions && auto_completing &&
			((e->key() & 0xFFF0) != 0x1020) &&
			(((e->key() & 0xFFF0) != 0x1010) || completed_before) &&
			(e->key() != Key_Delete))
		{
			getCursorPosition(&c_line, &c_col);
			QString *text = getString(c_line);
			int start_position = text->findRev(QRegExp("[^a-zA-Z0-9_.]"), c_col - text->length() - 1);
			c_length = c_col - start_position - 1;
			if (c_length >= 3)
			{
				c_head = text->mid(start_position + 1, c_length);
				c_results = wordCompletions->completionsFor(c_head);

				if (!c_results.isEmpty())
				{
					QString nextChar = QString("") + (*text)[c_col];
					QRegExp nextRegExp("[^a-zA-Z0-9_.]");
					completing = ((int) text->length() < c_col + 1) || (nextRegExp.match(nextChar) != -1);

					if (completing)
					{
						c_iterator = c_results.begin();
						start_iterator = c_results.end();
						do
							--start_iterator;
						while ((*start_iterator) == c_head);
						shift_used = false;

						QString word = (*c_iterator).mid(c_length);
						insert(word, true);

						++c_iterator;
					}
				}
			}
		}
	}

	emitCursorPosition();
}

void CompletingMultiLineEdit::mousePressEvent(QMouseEvent *e)
{
	leaveCompletionMode();
	QtMultiLineEdit::mousePressEvent(e);
	emitCursorPosition();
}

void CompletingMultiLineEdit::focusInEvent(QFocusEvent *e)
{
	QtMultiLineEdit::focusInEvent(e);
	emitCursorPosition();
}

void CompletingMultiLineEdit::emitCursorPosition()
{
	int new_line, new_column;
	getCursorPosition(&new_line, &new_column);
	if ((new_line != old_line) || (new_column != old_column))
	{
		old_line = new_line;
		old_column = new_column;
		emit cursorPositionChanged(new_line, new_column);
	}
}

void CompletingMultiLineEdit::setWordCompletions(WordCompletions *_wordCompletions)
{
	wordCompletions = _wordCompletions;
}

void CompletingMultiLineEdit::leaveCompletionMode()
{
	if (completing)
	{
		if (hasMarkedText())
			del();
		completing = false;
	}
}

void CompletingMultiLineEdit::setAutoCompleting(bool _auto_completing)
{
	auto_completing = _auto_completing;
}

#include "completing_multilineedit.moc"
