/*
 * Written by Volker Augustin, snowbird@bigfoot.de (03/2001)
 */

#include "ca_qt.h"
#include "interactive_close.h"
#include "main.h"
#include <qfontdialog.h>
#include <stdlib.h>

QString ApplicationWindow::defaultProperties[][2] = {
	{APP_FONT_NAME, "Courier"},
	{APP_FONT_SIZE, "12"},
	{APP_WORD_WRAP, WORD_WRAP_NORMAL},
	{APP_OUTPUT_WINDOW_POSITION, OWP_TOP},
	{APP_WINDOW_X, "50"},
	{APP_WINDOW_Y, "50"},
	{APP_WINDOW_WIDTH, "600"},
	{APP_WINDOW_HEIGHT, "400"},
	{APP_AUTO_COMPLETION, "0"},
	/*
	 * HACK
	 * {APP_AUTO_INDENTATION, "0"},
	 */
	{APP_TABWIDTH, "4"},
	{APP_FILE_TOOLBAR_ORIENTATION, QString::number(Qt::Horizontal)},
	{APP_FILE_TOOLBAR_DOCK, QString::number(QMainWindow::Top)},
	{APP_FILE_TOOLBAR_INDEX, "0"},
	{APP_FILE_TOOLBAR_NEWLINE, "0"},
	{APP_FILE_TOOLBAR_EXTRAOFFSET, "-1"},
	{APP_COCOA_TOOLBAR_ORIENTATION, QString::number(Qt::Horizontal)},
	{APP_COCOA_TOOLBAR_DOCK, QString::number(QMainWindow::Top)},
	{APP_COCOA_TOOLBAR_INDEX, "1"},
	{APP_COCOA_TOOLBAR_NEWLINE, "0"},
	{APP_COCOA_TOOLBAR_EXTRAOFFSET, "-1"},
	{APP_SHOW_INTERACTIVE_DOCUMENT_CLOSE_WARNING, "1"}
};

const QString& ApplicationWindow::getAppProperty(const QString& key)
{
	return properties.get(SECTION_APPLICATION_PROPERTIES)->get(key);
}

void ApplicationWindow::setAppProperty(const QString& key, const QString& value)
{
	properties.get(SECTION_APPLICATION_PROPERTIES)->set(key, value);
}

void ApplicationWindow::moveEvent( QMoveEvent *evt )
{
	QMainWindow::moveEvent(evt);
	if (!isMaximized())
	{
		QPoint position = evt->pos();
		setAppProperty(APP_WINDOW_X, QString::number(position.x()));
		setAppProperty(APP_WINDOW_Y, QString::number(position.y()));
	}
}

void ApplicationWindow::resizeEvent(QResizeEvent *evt)
{
	QMainWindow::resizeEvent(evt);
	if (!isMaximized())
	{
		QSize size = evt->size();
		int width = size.width();
		int height = size.height();
		setAppProperty(APP_WINDOW_WIDTH, QString::number(width));
		setAppProperty(APP_WINDOW_HEIGHT, QString::number(height));
	}
}

void ApplicationWindow::showPreferencesDialog()
{
	PreferencesDialog *dlg = new PreferencesDialog(this);
	connect(dlg, SIGNAL(addCompletionFiles(QStringList)), this, SLOT(loadCompletionFiles(QStringList)));
	connect(dlg, SIGNAL(loadCompletionFiles(QStringList)), this, SLOT(loadCompletionFiles(QStringList)));
	connect(dlg, SIGNAL(removeCompletionFiles(QStringList)), this, SLOT(removeCompletionFiles(QStringList)));
	connect(dlg, SIGNAL(unloadCompletionFiles(QStringList)), this, SLOT(unloadCompletionFiles(QStringList)));
	connect(this, SIGNAL(wordFileListUpdated(WordFileStateList)), dlg, SLOT(updateWordFileList(WordFileStateList)));
	emit wordFileListUpdated(wordFileStateList);
	dlg->show();
}

void ApplicationWindow::setWordWrap(int menuItemId)
{
	WordWrap *wrapType, *selectedWrapType;
	for (wrapType = wordWrapTypes.first(); wrapType != 0; wrapType = wordWrapTypes.next())
	{
		if (wrapType->menuIndex == menuItemId)
		{
			settingsMenu->setItemChecked(wrapType->menuIndex, true);
			selectedWrapType = wrapType;
		} else
			settingsMenu->setItemChecked(wrapType->menuIndex, false);
	}

	setAppProperty(APP_WORD_WRAP, selectedWrapType->type);
	wordWrapIndicator->setText(QString("WW: ") + selectedWrapType->type);
	output->selectWordWrap(selectedWrapType->type);
}

void ApplicationWindow::setAutoCompleting(bool on)
{
	settingsMenu->setItemChecked(autoCompletingId, on);
	setAppProperty(APP_AUTO_COMPLETION, on ? ENABLED : DISABLED);
	autoCompletionIndicator->setText(QString("AC: ") + (on ? "On" : "Off"));
	emit autoCompleting(on);
}

void ApplicationWindow::setAutoIndenting(bool on)
{
	settingsMenu->setItemChecked(autoIndentingId, on);
	/* HACK */
	if (getAppProperty(APP_AUTO_INDENTATION) != QString::null)
		setAppProperty(APP_AUTO_INDENTATION, on ? ENABLED : DISABLED);
	autoIndentationIndicator->setText(QString("AI: ") + (on ? "On" : "Off"));
	emit autoIndenting(on);
}

void ApplicationWindow::setOutputWindowPosition(int position)
{
	if (position == outputTopId)
	{
		settingsMenu->setItemChecked(outputTopId, true);
		settingsMenu->setItemChecked(outputBottomId, false);
		setAppProperty(APP_OUTPUT_WINDOW_POSITION, OWP_TOP);
		splitter->moveToFirst(output);
	} else
	{
		settingsMenu->setItemChecked(outputTopId, false);
		settingsMenu->setItemChecked(outputBottomId, true);
		setAppProperty(APP_OUTPUT_WINDOW_POSITION, OWP_BOTTOM);
		splitter->moveToLast(output);
	}
	splitter->refresh();
}

void ApplicationWindow::loadDefaultProperties()
{
	int n = sizeof(defaultProperties) / sizeof(defaultProperties[0]);
	PropertySection *section = properties.get(SECTION_APPLICATION_PROPERTIES);
	for (int i = 0; i < n; i++)
	{
		if (!section->contains(defaultProperties[i][0])
			|| (section->get(defaultProperties[i][0]) == QString::null))
		{
			setAppProperty(defaultProperties[i][0], defaultProperties[i][1]);
			cout << "Loaded default value for " << defaultProperties[i][0].latin1() << endl;
		}
	}
}

int ApplicationWindow::loadProperties()
{
	QString s;
	QFile propertyFile(propertyFileName);
	if (!propertyFile.open(IO_ReadOnly))
	{
		setStatus(QString("Can't open property file %1 for reading").arg(propertyFileName));
		return 1;
	}

	if (propertyFile.atEnd())
	{
		setStatus(QString("File %1 contains no data").arg(propertyFileName));
		return 1;
	}

	QTextStream in(&propertyFile);
	in >> properties;
	propertyFile.close();

	return 0;
}

int ApplicationWindow::saveProperties()
{
	QString s;
	QFile propertyFile(propertyFileName);
	if (!propertyFile.open(IO_WriteOnly))
	{
		setStatus(QString("Can't open property file %1 for writing").arg(propertyFileName));
		return 1;
	}

	QTextStream out(&propertyFile);
	out << properties;
	propertyFile.close();

	return 0;
}

void ApplicationWindow::selectFont(QFont f)
{
	currentFont = f;
	setAppProperty(APP_FONT_NAME, currentFont.family());
	setAppProperty(APP_FONT_SIZE, QString::number(currentFont.pointSize()));
	emit FontSelected(f);
}

void ApplicationWindow::selectFont()
{
	bool ok;

	int fontSize = getAppProperty(APP_FONT_SIZE).toInt(&ok);
	QFont initial;
	if (ok)
	        initial = QFont(getAppProperty(APP_FONT_NAME), fontSize);
	else
		initial = QFont(getAppProperty(APP_FONT_NAME));

	QFont f = QFontDialog::getFont(&ok, initial);
	if (ok)
		selectFont(f);
}

void ApplicationWindow::closeEvent( QCloseEvent* ce )
{
	InteractiveCloseDialog dlg(this);
	if (getAppProperty(APP_SHOW_INTERACTIVE_DOCUMENT_CLOSE_WARNING) == ENABLED)
	{
		if (dlg.exec() == QDialog::Rejected)
		{
			setAppProperty(APP_SHOW_INTERACTIVE_DOCUMENT_CLOSE_WARNING, dlg.showAgain() ? "1" : "0");

			output->engineInterrupted(false);

			/*
			 * HACK - the functionality to abort quitting the engine should be integrated into
			 * the class EngineController that will be implemented in the next XCoCoA version.
			 * For the time being we call ca_continue directly.
			 */
			ca_continue();

			return;
		} else
			setAppProperty(APP_SHOW_INTERACTIVE_DOCUMENT_CLOSE_WARNING, dlg.showAgain() ? "1" : "0");
	}

	QValueList<int> values = splitter->sizes();
	int topSize = values.first();
	int bottomSize = values.last();
	setAppProperty(APP_TOP_WINDOW_SIZE, QString::number(topSize));
	setAppProperty(APP_BOTTOM_WINDOW_SIZE, QString::number(bottomSize));
	setAppProperty(APP_FILE_TOOLBAR_ORIENTATION, QString::number(fileTools->orientation()));
	setAppProperty(APP_COCOA_TOOLBAR_ORIENTATION, QString::number(cocoaTools->orientation()));

	ToolBarDock toolbarDock;
	int toolbarIndex;
	bool toolbarNewLine;
	int toolbarExtraOffset;
	getLocation(fileTools, toolbarDock, toolbarIndex, toolbarNewLine, toolbarExtraOffset);
	setAppProperty(APP_FILE_TOOLBAR_DOCK, QString::number(toolbarDock));
	setAppProperty(APP_FILE_TOOLBAR_INDEX, QString::number(toolbarIndex));
	setAppProperty(APP_FILE_TOOLBAR_NEWLINE, QString::number(toolbarNewLine));
	setAppProperty(APP_FILE_TOOLBAR_EXTRAOFFSET, QString::number(toolbarExtraOffset));

	getLocation(cocoaTools, toolbarDock, toolbarIndex, toolbarNewLine, toolbarExtraOffset);
	setAppProperty(APP_COCOA_TOOLBAR_DOCK, QString::number(toolbarDock));
	setAppProperty(APP_COCOA_TOOLBAR_INDEX, QString::number(toolbarIndex));
	setAppProperty(APP_COCOA_TOOLBAR_NEWLINE, QString::number(toolbarNewLine));
	setAppProperty(APP_COCOA_TOOLBAR_EXTRAOFFSET, QString::number(toolbarExtraOffset));

	saveProperties();

	if (!output->isExecuting())
	{
		if (input->closeAllDocuments())
		{
			delete help;
			ce->accept();
		} else
			ce->ignore();
	} else
		exit(0);
}
