/*
 * Written by Arndt Wills <Arndt.Wills@stud.uni-regensburg.de> (02/2001)
 */

#include <qaccel.h>
#include <qfiledialog.h>
#include <qfileinfo.h>
#include <qiodevice.h>
#include <qmessagebox.h>
#include <qtextstream.h>

#include <qobjectlist.h>
#include <qobjectdict.h>
#include <qtabbar.h>
#include <qapplication.h>
#include <qwidgetstack.h>
#include <qbitmap.h>
#include "simple_tabwidget.h"

class SimpleTabWidgetData
{
public:
	SimpleTabWidgetData()
		: tabs(0), stack(0), dirty( TRUE ), pos( SimpleTabWidget::Top ), shape( SimpleTabWidget::Rounded ) {}
	~SimpleTabWidgetData(){}
	SimpleTabBar* tabs;
	QWidgetStack* stack;
	bool dirty;
	SimpleTabWidget::TabPosition pos;
	SimpleTabWidget::TabShape shape;
};

SimpleTabWidget::SimpleTabWidget(QWidget *parent, const char *name)
	: QWidget(parent, name)
{
	d = new SimpleTabWidgetData;

	d->stack = new QWidgetStack(this, "tab pages");
	d->stack->installEventFilter(this);
	setTabBar(new SimpleTabBar(this, "tab control"));

	d->stack->setFrameStyle(QFrame::StyledPanel | QFrame::Sunken);

	setSizePolicy(QSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding));
	setFocusPolicy(TabFocus);
	setFocusProxy(d->tabs);

	QAccel *a = new QAccel(this);
	a->connectItem(a->insertItem(ALT + Key_Plus), this, SLOT(nextPage()));
	a->connectItem(a->insertItem(ALT + Key_Minus), this, SLOT(previousPage()));
}

void SimpleTabWidget::previousPage()
{
	int id = d->tabs->previousTab();
	d->tabs->setCurrentTab(id);
	showTab(id);
}

void SimpleTabWidget::nextPage()
{
	int id = d->tabs->nextTab();
	d->tabs->setCurrentTab(id);
	showTab(id);
}

SimpleTabWidget::~SimpleTabWidget()
{
	delete d;
}

void SimpleTabWidget::addTab(QWidget *child, const QString &label)
{
	QTab * t = new QTab();
	CHECK_PTR(t);
	t->label = label;
	addTab(child, t);
}

void SimpleTabWidget::addTab(QWidget *child, const QIconSet& iconset, const QString &label)
{
	QTab * t = new QTab();
	CHECK_PTR(t);
	t->label = label;
	t->iconset = new QIconSet(iconset);
	addTab(child, t);
}

void SimpleTabWidget::addTab(QWidget *child, QTab* tab)
{
	tab->enabled = TRUE;
	int id = d->tabs->addTab(tab);
	d->stack->addWidget(child, id);
	setUpLayout();
}

void SimpleTabWidget::insertTab(QWidget *child, const QString &label, int index)
{
	QTab * t = new QTab();
	CHECK_PTR(t);
	t->label = label;
	insertTab(child, t, index);
}

void SimpleTabWidget::insertTab(QWidget *child, const QIconSet& iconset, const QString &label, int index)
{
	QTab * t = new QTab();
	CHECK_PTR(t);
	t->label = label;
	t->iconset = new QIconSet(iconset);
	insertTab(child, t, index);
}

void SimpleTabWidget::insertTab(QWidget *child, QTab* tab, int index)
{
	tab->enabled = TRUE;
	int id = d->tabs->insertTab(tab, index);
	d->stack->addWidget(child, id);
	setUpLayout();
}

void SimpleTabWidget::changeTab(QWidget *w, const QString &label)
{
	 //#### accelerators
	int id = d->stack->id(w);
	if (id < 0)
		return;
	QTab* t = d->tabs->tab(id);
	if (!t)
		return;
	t->label = label;

	int ct = d->tabs->currentTab();
        d->tabs->changeTab(t);
	d->tabs->layoutTabs();

	d->tabs->blockSignals(TRUE);
	d->tabs->setCurrentTab(0);
	d->tabs->setCurrentTab(ct);
	d->tabs->blockSignals(FALSE);

	d->tabs->update();
	setUpLayout();
}

void SimpleTabWidget::changeTab(QWidget *w, const QIconSet& iconset, const QString &label)
{
	//#### accelerators
	int id = d->stack->id(w);
	if (id < 0)
		return;
	QTab* t = d->tabs->tab(id);
	if (!t)
		return;
	if (t->iconset)
		delete t->iconset;
	t->label = label;
	t->iconset = new QIconSet(iconset);

	int ct = d->tabs->currentTab();
	d->tabs->changeTab(t);
	d->tabs->layoutTabs();

	d->tabs->blockSignals(TRUE);
	d->tabs->setCurrentTab(0);
	d->tabs->setCurrentTab(ct);
	d->tabs->blockSignals(FALSE);

	d->tabs->update();
	setUpLayout();
}

/*obsolete?*/
bool SimpleTabWidget::isTabEnabled(QWidget* w) const
{
	int id = d->stack->id(w);
	if (id >= 0)
		return w->isEnabled();
	else
		return FALSE;
}

/*obsolete?*/
void SimpleTabWidget::setTabEnabled(QWidget* w, bool enable)
{
	int id = d->stack->id(w);
	if (id >= 0)
	{
		w->setEnabled(enable);
		d->tabs->setTabEnabled(id, enable);
	}
}

void SimpleTabWidget::showPage(QWidget * w)
{
	int id = d->stack->id(w);
	if (id >= 0)
	{
		d->stack->raiseWidget(w);
		d->tabs->setCurrentTab(id);
	}
}

void SimpleTabWidget::removePage(QWidget * w)
{
	int id = d->stack->id(w);
	if (id >= 0)
	{
		bool current = (id == d->tabs->currentTab());
		int next = -1;
		if (current)
		{
			next = d->tabs->nextTab();
			if (next == 0)
				next = d->tabs->previousTab();
		}
		d->tabs->setTabEnabled(id, FALSE);
		d->stack->removeWidget(w);
		d->tabs->removeTab(d->tabs->tab(id));
		if (current)
			setCurrentPage(next);
		setUpLayout();
	}
}

/*obsolete?*/
QString SimpleTabWidget::tabLabel(QWidget * w)
{
	QTab * t = d->tabs->tab(d->stack->id(w));
	return t ? t->label : QString::null;
}

QWidget * SimpleTabWidget::currentPage() const
{
	return d->stack->visibleWidget();
}

int SimpleTabWidget::currentPageIndex() const
{
	return d->tabs->currentTab();
}

void SimpleTabWidget::setCurrentPage(int id)
{
	d->tabs->setCurrentTab(id);
	showTab(id);
}

QSize SimpleTabWidget::sizeHint() const
{
	QSize s(d->stack->sizeHint());
	QSize t(d->tabs->sizeHint());
	return QSize(QMAX(s.width(), t.width()), s.height() + t.height());
}

QSize SimpleTabWidget::minimumSizeHint() const
{
	QSize s(d->stack->minimumSizeHint());
	QSize t(d->tabs->minimumSizeHint());
	return QSize(QMAX(s.width(), t.width()), s.height() + t.height());
}

SimpleTabWidget::TabPosition SimpleTabWidget::tabPosition() const
{
	return d->pos;
}

void SimpleTabWidget::setTabPosition(TabPosition pos)
{
	if (d->pos == pos)
		return;
	d->pos = pos;
	if (d->tabs->shape() == SimpleTabBar::TriangularAbove || d->tabs->shape() == SimpleTabBar::TriangularBelow )
	{
		if (pos == Bottom)
			d->tabs->setShape(SimpleTabBar::TriangularBelow);
		else
			d->tabs->setShape(SimpleTabBar::TriangularAbove);
	} else {
		if (pos == Bottom)
			d->tabs->setShape(SimpleTabBar::RoundedBelow);
		else
			d->tabs->setShape(SimpleTabBar::RoundedAbove);
	}
	d->tabs->layoutTabs();
	setUpLayout();
}

SimpleTabWidget::TabShape SimpleTabWidget::tabShape() const
{
	return d->shape;
}

void SimpleTabWidget::setTabShape(TabShape s)
{
	if (d->shape == s)
		return;
	d->shape = s;
	if (d->pos == Top)
	{
		if (s == Rounded)
			d->tabs->setShape(SimpleTabBar::RoundedAbove);
		else
			d->tabs->setShape(SimpleTabBar::TriangularAbove);
	} else {
		if (s == Rounded)
			d->tabs->setShape(SimpleTabBar::RoundedBelow);
		else
			d->tabs->setShape(SimpleTabBar::TriangularBelow);
	}
	d->tabs->layoutTabs();
	setUpLayout();
}

int SimpleTabWidget::margin() const
{
	return d->stack->margin();
}

void SimpleTabWidget::setMargin(int w)
{
	d->stack->setMargin(w);
	setUpLayout();
}

/*obsolete?*/
QSizePolicy SimpleTabWidget::sizePolicy() const
{
	//### removeme 3.0
	return QWidget::sizePolicy();
}

void SimpleTabWidget::showEvent(QShowEvent *)
{
	setUpLayout(TRUE);
}

void SimpleTabWidget::resizeEvent(QResizeEvent *)
{
	setUpLayout();
}

void SimpleTabWidget::setTabBar(SimpleTabBar* tb)
{
	if (tb->parentWidget() != this)
		tb->reparent(this, QPoint(0,0), TRUE);
	delete d->tabs;
	d->tabs = tb;
	connect(d->tabs, SIGNAL(selected(int)), this, SLOT(showTab(int)));
	setUpLayout();
}

SimpleTabBar* SimpleTabWidget::tabBar() const
{
	return d->tabs;
}

void SimpleTabWidget::styleChange(QStyle& old)
{
	d->stack->setLineWidth(style().defaultFrameWidth());
	setUpLayout();
	QWidget::styleChange(old);
}

void SimpleTabWidget::updateMask()
{
	if (!autoMask())
		return;

	QRect r;
	QRegion reg(r);
	reg += QRegion(d->tabs->geometry());
	reg += QRegion(d->stack->geometry());
	setMask(reg);
}

bool SimpleTabWidget::eventFilter(QObject *o, QEvent * e)
{
	if (o == d->stack && e->type() == QEvent::ChildRemoved
		&& ((QChildEvent*) e)->child()->isWidgetType())
	{
		removePage((QWidget*)((QChildEvent*) e)->child());
		return TRUE;
	}
	return FALSE;
}

void SimpleTabWidget::showTab(int i)
{
	if (d->stack->widget(i))
	{
		d->stack->raiseWidget(i);
		emit selected(d->tabs->tab(i)->label);
		emit currentChanged(d->stack->widget(i));
	}
}

void SimpleTabWidget::setUpLayout(bool onlyCheck)
{
	if (onlyCheck && !d->dirty)
		return; // nothing to do

	if (!isVisible())
	{
		d->dirty = TRUE;
		return; // we'll do it later
	}
	QSize t(d->tabs->sizeHint());
	if (t.width() > width())
		t.setWidth(width());
	int lw = d->stack->lineWidth();
	if (d->pos == Bottom)
	{
		d->tabs->setGeometry(QMAX(0, lw-2) + 1, height() - t.height() - lw, t.width(), t.height());
		d->stack->setGeometry(0, 0, width(), height() - t.height() + QMAX(0, lw - 2) - 1);
	}
	else
	{ // Top
		d->tabs->setGeometry(QMAX(0, lw - 2), 0, t.width(), t.height());
		d->stack->setGeometry(0, t.height(), width(), height() - t.height() + QMAX(0, lw - 2));
	}

	d->dirty = FALSE;
	if (!onlyCheck)
		update();
	if (autoMask())
		updateMask();
}

#include "simple_tabwidget.moc"
