#include "net_lib.h"

#define MIN(x,y) ((x) < (y) ? (x) : (y))

extern int errno; /* Needed to catch system calls errors */
extern int h_errno; /* Needed to catch resolver errors */

int error(char* prefix)
{
	fprintf(stderr, "Error: %s: %s\n", prefix, strerror(errno));
	switch(errno)
	{
		case EBADF:
		case ENOTSOCK:
		case EINVAL:
		case ENOTCONN: return PARAM_ERR;

		case ETIMEDOUT:
		case ENETUNREACH:
		case ECONNREFUSED: return SERV_ERR;
		case EADDRINUSE:
		case EIO:
		case EPROTONOSUPPORT:
		case EAFNOSUPPORT:
		case EINTR: return OS_ERR;

		case ENFILE:
		case EMFILE:
		case ENOBUFS:
		case ENOMEM: return MEM_ERR;

		case EMSGSIZE: return SIZE_ERR;

		case EPIPE: return CONN_ERR;

		case EACCES:
		case EPERM: return PERM_ERR;

		default: return UNKN_ERR;
	}

}

void error_str(char* prefix, char* error)
{
	fprintf(stderr, "Error: %s: %s\n", prefix, error);
	return;
}

int cocoa_bind(struct sock_t *sock, port_t port_number)
{
	int retval;

	/* Address where the server will be listening, as sin_port we'll set 'port'
	 * for sin_addr we'll use INADDR_ANY that means 'listen on all active interfaces'
	 * For IPv6 the generic sockaddr_storage should be used instead.
	 */
	struct sockaddr_in server_listen_address;

	/* Socket creation, IP with connection oriented transpot layer, protocol 0 (TCP)
	 * For IPv6 PF_INET6 should be used instead, for dual stack system more work is 
	 * needed, currently different operating systems behave differently */
	sock->fd = socket(PF_INET, SOCK_STREAM, 0);
	if (sock->fd < 0)
		return error("cocoa_bind");

	server_listen_address.sin_family = AF_INET;
	server_listen_address.sin_port = htons(port_number);
	server_listen_address.sin_addr.s_addr = (int)htonl(INADDR_ANY);

	/* Bind the socket sock to the address server_listen_address */
	retval = bind(sock->fd, (struct sockaddr*)&server_listen_address, sizeof(server_listen_address));
	if (retval < 0) 
		return error("cocoa_bind");
	
	listen(sock->fd, WAIT_QUEUE_LEN);

	sock->is_alive = 1;

	return sock->fd;
}

int cocoa_accept(struct sock_t *sock, void (*handler)(struct sock_t*))
{
	struct sock_t client_sock; /* Socket created by accept() to talk to clients */
	int pid; /* Child process PID */
	struct sockaddr_in client_address; /* Client address, for IPv6 use sockaddr_storage */
	int addr_size; /* Client address length */

	addr_size = sizeof(client_address);
	
	init_sock(&client_sock);
	
	/* Accepts a connection, if the queue is empty the process sleeps until some
	 * client tries to connect.
	 */
	client_sock.fd = accept(sock->fd, (struct sockaddr*)&client_address, &addr_size);
	if (client_sock.fd < 0) 
		return error("cocoa_accept");
	/* Fork a child process that will have 0 as return value */
	pid = fork();
	if (pid == 0) 
	{
		client_sock.is_alive = 1;
#ifdef DEBUG
		fprintf(stderr, "Child process spawned\n");
#endif
		handler(&client_sock);
		cocoa_close(&client_sock);
#ifdef DEBUG
		fprintf(stderr, "Child process finished\n");
#endif
		exit(0);
	}
	cocoa_close(&client_sock);

	return pid;
}

int cocoa_connect(struct sock_t *sock, const char hostname[], port_t port_number)
{
	struct sockaddr_in server; /* IPv6 requires sockaddr_storage */
	struct hostent *hostinfo; /* IPv6 needs use of gethostinfo() */
	int retval;

	/* IPv6 wants PF_INET6 */
	sock->fd = socket(PF_INET, SOCK_STREAM, 0);
	if (sock->fd < 0)
		return error("cocoa_connect");

	server.sin_family = AF_INET; /* IPv6 AF_INET6 */
	server.sin_port = htons(port_number);
	
	/* We need to translate the hostname in an IP address 
	 * gethostbyname() queries /etc/hosts, DNS, NIS, ... in the correct
	 * order. IPv6 needs gethostinfo(), not available on older systems
	 */
	hostinfo = gethostbyname(hostname);
	if (hostinfo == NULL) /* DNS resolver needs its own error checking */
	{
		fprintf(stderr, "Error: cocoa_connect: %s\n", hstrerror(h_errno));
		switch(h_errno)
		{
			case HOST_NOT_FOUND: return PARAM_ERR;
			case NO_ADDRESS:
			case NO_RECOVERY: return DNS_ERR;
			case TRY_AGAIN: return DNS_WARN;
			default: return UNKN_ERR;
		}
	}
	
	server.sin_addr = *(struct in_addr *)hostinfo->h_addr;
	
	/* Connect using the socket sock to the server specified by server struct */
	retval = connect(sock->fd, (struct sockaddr*)&server, sizeof(server));
	if (retval < 0)
		return error("cocoa_connect");

	sock->is_alive = 1;

	return sock->fd;
}

int cocoa_send(struct sock_t *sock, void *buffer, size_t buffer_len)
{
	int retval, num_chunks, i;
	void *pos;
	size_t chunk_size;

        if (!sock->is_alive) { return error("cocoa_send: socket has already been closed"); }
	if (buffer_len > 0xFFFFFFFF)
	{
		error_str("cocoa_send", "cannot send more than 4GB in a single call");
		return SIZE_ERR;
	}

	num_chunks = buffer_len/CHUNK_LEN + 1;

#ifdef DEBUG
	printf("For %d byte(s), %d chunks are needed with a chunk_len of %d\n", buffer_len, num_chunks, CHUNK_LEN);
#endif

	pos = buffer;
	for(i=0; i<num_chunks; i++) {
		chunk_size = MIN(CHUNK_LEN, buffer+buffer_len-pos);
#ifdef DEBUG
		fprintf(stderr, "Sending %d byte(s) of data\n", chunk_size);
#endif
		retval = send(sock->fd, pos, chunk_size, 0);
		if (retval == -1)
			return error("cocoa_send");
		pos += chunk_size;
	}

	return 0;
}

int cocoa_receive(struct sock_t *sock)
{
	int retval=0;

	if (!sock->is_alive)
		return CONN_ERR;

	retval = recv(sock->fd, sock->recv_buffer, RECV_BUFFER_LEN, 0);
	if (retval == -1)
		return CONN_ERR;
	else if (retval == 0) /* the other end closed the connection */
	{
          sock->data_len = 0;
          return CONN_ERR;
	}
#ifdef DEBUG
	fprintf(stderr, "Read %d bytes from socket\n", retval);
#endif

	/* All went well, so set data_len to the correct value */
	sock->data_len = retval;

	return 0;
}

int cocoa_sock_getc(struct sock_t *sock, int* c)
{
	int retval = 0;

        if (!sock->is_alive) { *c = EOF; return CONN_ERR; }
	if (sock->pos >= sock->data_len)
	{
		retval = cocoa_receive(sock);
		if (retval == CONN_ERR)
		{
			*c = EOF;
			cocoa_close(sock);
			return retval;
		}
		sock->pos = 0;
	}

	*c = sock->recv_buffer[sock->pos];
	sock->pos++;
	return retval;
}

int cocoa_close(struct sock_t *sock)
{
	int retval;
	
	if (sock == 0 || !sock->is_alive)
		return 0;

	sock->is_alive = 0;
	retval = close(sock->fd);
	if (retval < 0)
		return error("cocoa_close");

	sock->fd = -1;
	return 0;
}

void init_sock(struct sock_t *sock)
{
	sock->fd = -1;
	memset(sock->recv_buffer, 0, RECV_BUFFER_LEN);
	sock->data_len = 0;
	sock->pos = 0;
	sock->is_alive = 0;

	return;
}

int is_connected(struct sock_t *sock)
{
	return sock->is_alive;
}

