/* This library provides a complete interface to CoCoA to do network
 * communication using sockets.
 *
 * It is ready to be ported to IPv6 eventually. Comments tagged 'IPv6'
 * will be found where code needs modification (in net_lib.c).
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#define OS_ERR		-1 /* Operating system error, nothing can be done */
#define MEM_ERR		-2 /* Out of memory (RAM or hardware buffers) */
#define PERM_ERR	-3 /* Permission denied errors */
#define UNKN_ERR	-4 /* Unknown error, unexpected or not documented */
#define PARAM_ERR	-5 /* Wrong parameters */
#define DNS_ERR		-6 /* DNS server errors */
#define SERV_ERR	-7 /* trying to connect to someone who is not responding or unreacheable */
#define SIZE_ERR	-8 /* Message was too big for network layer */
#define CONN_ERR	-9 /* Connection was dropped by a peer without notification */
#define VERS_ERR	-10 /* Version mismatch between client and server */

#define DNS_WARN	-100 /* The DNS layer gave an error, but thinks that later things'll be better */
#define OS_WARN		-101 /* The operating system gave some warning that can be ignored */

#ifdef __cplusplus
extern "C" {
#endif

/* This is right for alpha, Intel (32 and 64 bits) and all other architectures
 * for which there is a Linux kernel port
 */
typedef unsigned short port_t;

/* Queue length before connection requests start to be discarded */
#define WAIT_QUEUE_LEN 10

/* Receive buffer size, it is chosen to be half of CHUNK_LEN to give some more
 * performance, less calls to cocoa_receive should block witing for new data.
 */
#define RECV_BUFFER_LEN (2<<13)

/* cocoa_send divides the messaage in chunks of this size */
#define CHUNK_LEN (2<<14)

/* Socket struct */
struct sock_t
{
	int fd;
	char recv_buffer[RECV_BUFFER_LEN];
	int data_len;
	int pos;
	int is_alive;
};

/* Fills sock with meaningful values */
void init_sock(struct sock_t *sock);

/* Returns true if the socket is connected, false otherwise */
int is_connected(struct sock_t *sock);

/* Prepares a socket to receive incoming connections on portport_number. 
 * It'll make the socket listen on all the local interfaces, 
 * finer control shouldn't be needed to CoCoA.
 *
 * On error returns a value < 0, else a connection descriptor is returned, 
 * to be used in cocoa_accept calls.
 */
int cocoa_bind(struct sock_t *sock, port_t port_number);

/* Accepts a connection on the socket identified by sockfd and forks a process
 * to serve the new request. The child process calls handler with the new
 * socket as parameter, the parent process returns from the call to
 * cocoa_accept.
 *
 * If no connection request is available the process blocks waiting.
 *
 * On error returns a value < 0, on normal operation the PID of the child
 * process is returned instead.
 */ 
int cocoa_accept(struct sock_t *sock, void (*handler)(struct sock_t*));

/* Makes a connection to hostname on port port_number.
 *
 * On error returns a value < 0, else a connection descriptor is returned, 
 * to be used in coca_send and cocoa_receive calls
 */
int cocoa_connect(struct sock_t *sock, const char hostname[], port_t port_number);

/* Sends the contents of buffer to the other end of sockfd, exactly buffer_len
 * bytes of data will be sent.
 *
 * On error returns a value < 0, on normal operation 0 is returned instead.
 */
int cocoa_send(struct sock_t *sock, void *buffer, size_t buffer_len);

/* Receives data from the socket sockfd, the function fills the 
 * cocoa_receive_buffer buffer with noe data up to RECV_BUFFER_LEN.
 * The length in bytes of data read is written in data_len.
 *
 * On error returns a value < 0, on normal operation 0 is returned instead.
 */
int cocoa_receive(struct sock_t *sock);

/* Frontend to cocoa_receive that returns one byte (cast to char) of data read
 * from the socket sockfd.
 * It is buffered, so if you use this function, don't mix calls with cocoa_receive
 * as it will not return the remaining data in the cocoa_sock_getc buffer before
 * reading from the socket.
 * 
 * On error returns a value < 0, on normal operation 0 is returned instead.
 */
int cocoa_sock_getc(struct sock_t *sock, int* c);

/* Closes and frees all resources associated with sockfd, should be called
 * when the connection is no more needed or the server should stop listening
 * for incoming connections.
 */
int cocoa_close(struct sock_t *sock);

#ifdef __cplusplus
}
#endif
