
#include "DUPZroot_bound.h"
#include "mpz_log.h"
#include <math.h>
/* Must define our own log for integers -- some compilers get it wrong. */
#include "logi.h"

/* This finds an upper bound on the magnitude of the roots of f using    */
/* Knuth's bound -- see "Seminumerical Algorithms" ex. 4.6.2-20 (1969 ed) */

static double DUPZroot_bound_knuth(const DUPZ f)
{
  int df, i;
  double ans, loglcf, logfi;

  df = DUPZdeg(f);
  loglcf = mpz_log(f->coeffs[df]);
  ans = -loglcf;
  for (i = 0; i < df; i++)
  {
    if (mpz_sgn(f->coeffs[i]) == 0) continue;
    logfi = mpz_log(f->coeffs[i]) - loglcf;
    if (ans < logfi/(df-i)) ans = logfi/(df-i);
  }
  return ans + logi(2);
}


/* This corresponds to exercise 4.6.2-19 in "Seminumerical Algorithms" (1969 ed) */

static double DUPZroot_bound_zassenhaus(const DUPZ f)
{
  int df, i;
  double ans, loglcf, term, log_binomial;
 
  df = DUPZdeg(f);
  loglcf = mpz_log(f->coeffs[df]);
  ans = - loglcf - df*logi(2);
  log_binomial = 0.0;
  for (i = df-1; i >= 0; i--)
  {
    log_binomial += logi(i+1) - logi(df-i);
    if (mpz_sgn(f->coeffs[i]) == 0) continue;
    term = mpz_log(f->coeffs[i]) - loglcf - log_binomial;
    if (ans < term/(df-i)) ans = term/(df-i);
  }
  return ans - log(exp(logi(2)/df)-1);

}


/* Compute both Knuth's root bound, and Zassenhaus's; and return the smaller */
double DUPZroot_bound(const DUPZ f)
{
  double K, Z;
  K = DUPZroot_bound_knuth(f);
  Z = DUPZroot_bound_zassenhaus(f);
  if (K < Z) return K;
  return Z;
}
