/***************************************************************************/
/* Work still to do: projective->affine reduction when appropriate         */
/*                   handle rational points directly (BM_check_final)      */
/*                                                   (reduce_mod_p)        */
/*                   remove edge_pp condition???                           */
/*                   act sensibly when BM_check_final needs more accuracy  */
/*                   last useful pp in projective modular case             */
/*                   dynamic use of degree compatibility as far as poss    */
/*                                                                         */
/***************************************************************************/

/* These 3 includes are just for development... */
#include "jalloc.h"
#include <stdio.h>

#include <stddef.h>
#include "mpz_alias.h"
#include "GBM.h"
#include "PPstream.h"
#include "mpq_to_FFelem.h"

#include "primes.h"
#include "FF.h"
#include "DUPFF.h"
#include <gmp.h>
#include "Qsolve.h"

/***************************************************************************/
/* Two global flags indicating whether we're in affine/projective case     */
/* At any one time exactly one of these should be true: see BM_affine      */
/* and BM_projective below.                                                */

static int affine_case = 0;
static int projective_case = 0;


/***************************************************************************/
/* Test whether one power product is the "parent" of another.              */
/* Result is -1 if parent is not the parent of child,                      */
/* otherwise the index of the indeterminate relating parent to child.      */

static int pp_parent(int nvars, int *parent, int *child)
{
  int i, j;

  for (i=0; i < nvars; i++)
    if (parent[i] != 0 || child[i] != 0) break;
  if (i == nvars) return -1;
  if (parent[i] != child[i]-1) return -1;
  for (j=i+1; j < nvars; j++) if (parent[j] != child[j]) return -1;
  return i;
}


/***************************************************************************/
/* Compute a new row of the GBM matrix by "multiplying" a previous row.    */
/* The new power product must already be in this->pp[this->npp].           */
/* If pp = 1 then this is a special case.                                  */


static void make_new_row(GBM this)
{
  int i, j, prev;
  int x = 0; /* assignment to keep compiler quiet */
  int npp = this->npp;
  FFelem *v, *w;

  if (npp == 0)
  {
    v = this->image[npp];
    for (i=0; i < this->nideals; i++)
    {
      for (j=0; j < this->quotdim[i]; j++)
        v[j] = this->one[i][j];
      v += this->quotdim[i];
    }
    return;
  }

  /* This section is pretty inefficient */
  for (prev=0; prev < npp; prev++)
    if ((x = pp_parent(this->nvars, this->pp[prev], this->pp[npp])) >= 0) break;
  v = this->image[prev];
  w = this->image[npp];
  for (i=0; i < this->nideals; i++)
  {
    FFmat_mul_vec(w, this->Mmult[i][x], v);
    v += this->quotdim[i];
    w += this->quotdim[i];
  }
}


/***************************************************************************/
/* Constructor and destructor for the structure upon which GBM_mod_p acts. */
/* Initially we do not know how big to make the arrays, so we guess, and   */
/* offer the function BM_grow if our guess was too small.                  */
/* We only ever create one of these structures, and the fields pp, issep,  */
/* and GBsize keep their values between successive calls to BM_mod_p.      */
/* Furthermore, the array "det" is used only in the projective case (well, */
/* the affine case uses just det[0]); similarly projective_sep is use only */
/* in the projective case (set by the function fill_projective_sep, q.v.). */
/* npp_max tells us how big the arrays GB, issep, M, and pp are, while npp */
/* tells us how many of the elements are actually used: indices 0..(npp-1).*/

static GBM GBM_ctor(int nvars, int nideals, const int *quotdim, /*const*/ FFelem **one, /*const*/ FFmat **Mmult, int (*cmp)(const void*,const void*))
{
  int i;
  int npp_max, ncols;
  GBM this;

  this = (GBM)MALLOC(sizeof(struct GeneralizedBuchbergerMoeller_struct));
  ncols = 0;
  for (i=0; i < nideals; i++) ncols += quotdim[i];
  this->ncols = ncols;
  this->nseps = 0;
  npp_max = 2*ncols;  /* ideally should be # of PPs needed to get the answer */
  this->nideals = nideals;
  this->quotdim = quotdim;
  this->one = one;
  this->Mmult = Mmult;
  this->nvars = nvars;
  this->npp_max = npp_max;
  this->npp = 0;
  this->GB = (int*)MALLOC(npp_max*sizeof(int));
  this->sep = (int*)MALLOC(ncols*sizeof(int));
  /* Clear all separators */
  for (i=0; i < ncols; i++) this->sep[i] = -1;
  this->pp = (int**)MALLOC(npp_max*sizeof(int*));
  this->image = (FFelem**)MALLOC(npp_max*sizeof(FFelem*));
  this->M = (FFelem**)MALLOC(npp_max*sizeof(FFelem*));
  this->cmp = cmp;
  for (i=0; i < npp_max; i++)
  {
    this->pp[i] = NULL; /* space is created by PPstream_next */
    this->image[i] = (FFelem*)MALLOC(ncols*sizeof(FFelem));
    this->M[i] = (FFelem*)MALLOC(2*ncols*sizeof(FFelem));
  }
  this->GBsize = 0;
  return this;
}


void GBM_dtor(GBM this)
{
  int i;
  int npp_max = this->npp_max;

  for (i=0; i < npp_max; i++) { FREE(this->pp[i]); FREE(this->M[i]); FREE(this->image[i]); }
  FREE(this->image);
  FREE(this->M);
  FREE(this->pp);
  FREE(this->sep);
  FREE(this->GB);
  FREE(this);
}

/* For simplicity we merely double the maximum number of power products  */
/* which the structure can handle.  Repeated calls to BM_grow repeatedly */
/* double the space available.                                           */

static void GBM_grow(GBM this)
{
  int i;
  int npp_max = this->npp_max;
  
  this->npp_max = 2*npp_max;
  this->GB = (int*)REALLOC(this->GB, 2*npp_max*sizeof(int));
  this->pp = (int**)REALLOC(this->pp, 2*npp_max*sizeof(int*));
  this->M = (FFelem**)REALLOC(this->M, 2*npp_max*sizeof(FFelem*));
  this->image = (FFelem**)REALLOC(this->image, 2*npp_max*sizeof(FFelem*));
  for (i=npp_max; i < 2*npp_max; i++)
  {
    this->pp[i] = NULL;
    this->M[i] = (FFelem*)MALLOC(2*this->ncols*sizeof(FFelem));
    this->image[i] = (FFelem*)MALLOC(this->ncols*sizeof(FFelem));
  }
}

/***************************************************************************/
/* Check for linear dependence of the last row in GBM->M on earlier rows.  */
/* This function typically consumes almost all the execution time (in the  */
/* call to DUPFFshift_add_raw).                                            */
/* Note that the matrix GBM->M has 2*npoints columns: the first npoints cols*/
/* contain the row-reduced values of the power product, the last npoints   */
/* cols contain the linear combination of separator PPs which was used to  */
/* reduce that row.  As side effect, this procedure accumulates a sort of  */
/* "determinant" in the third arg -- this will be the determinant of the   */
/* original separator rows if a full set of separators is found.           */
/* The value returned is -1 if the row reduced to zero, otherwise it is the*/
/* index of the first column with a non-zero entry (from 0 up to npoints-1)*/
/* and the row has been scaled so that this entry is 1.                    */

static int reduce_new_row(GBM this)
{
  int pivot, j;
  FFelem *row, c;
  FFelem *v = this->M[this->npp];
  int ncols = this->ncols;
  int npp = this->npp;
  int *sep = this->sep;

  /* copy new image into the reduction matrix */
  for (j=0; j < ncols; j++)
  {
    v[j] = this->image[npp][j];
    v[j+ncols] = 0;
  }

  for (j=0; j < ncols; j++)
  {
    if (v[j] == 0 || sep[j] < 0) continue;
    row = this->M[sep[j]];
    c = CurrentFF.prime - v[j];
    DUPFFshift_add_raw(&v[j], &row[j], &row[2*ncols-1], c);
  }

  for (j=0; j < ncols; j++)
    if (v[j] != 0) break;
  if (j == ncols) { this->GB[this->GBsize++] = npp; return 1;}

  /* Got a new "separator", rescale it, and reduce previous separators by it */
  this->nseps++;
  pivot = j;
  sep[pivot] = npp;
  v[pivot+ncols] = 1;
  c = FFdiv(1, v[pivot]);
  for (j=0; j < 2*ncols; j++) v[j] = FFmul(v[j], c);
  for (j=0; j < ncols; j++)
  {
    if (sep[j] < 0) continue;
    if (j == pivot) continue; /* don't reduce self! */
    row = this->M[sep[j]];
    if (row[pivot] == 0) continue;
    c = CurrentFF.prime - row[pivot];
    DUPFFshift_add_raw(&row[0], &v[0], &v[2*ncols-1], c);
  }

  return 0;
}


/***************************************************************************/
/* The Generalized Buchberger-Moeller algorithm (modulo p) for PP orderings*/
/* which are NOT degree compatible.  It is valid also for degree compatible*/
/* orderings but then GBM_mod_p2 should be faster, perhaps by quite a lot. */

static void add_new_pp(GBM this, int* pp)
{
  int npp = this->npp;

  if (npp == this->npp_max) { GBM_grow(this); }
  this->pp[npp] = pp;
}

static void GBM_mod_p1(GBM this)
{
  int isGB;
  int nvars = this->nvars;
  int *pp;
  PPstream PPs;
  
  this->maxdeg = 0;
  PPs = PPstream_ctor(nvars, this->cmp);
  while ((pp = PPstream_next(PPs)))
  {
    add_new_pp(this, pp);
    make_new_row(this);
    isGB = reduce_new_row(this);
    if (isGB) PPstream_avoid_last(PPs);
    this->npp++;
  }
  PPstream_dtor(PPs);
  return;
}


/***************************************************************************/
/*  T H I S   I S   T H E   P U B L I C   I N T E R F A C E                */
/***************************************************************************/

GBM GBM_affine_mod_p(int nvars, int nideals, const int *quotdim, /*const*/ FFelem **one, /*const*/ FFmat **Mmult, int (*cmp)(const void*, const void*))
{
  GBM ans;

  affine_case = 1;
  projective_case = 0;
  ans = GBM_ctor(nvars, nideals, quotdim, one, Mmult, cmp);
  GBM_mod_p1(ans);
  return ans;
}


void GBM_affine(GBM *trace_out, Qmat *GBtails_out, int nvars, int nideals, const int *quotdim, mpq_t **one, Qmat **Mmult, int (*cmp)(const void*, const void*))
{
  int i, j, k, OK, ncols, nseps, **pp_col, s, t, GBsize;
  GBM trace;
  int x = 0;    /* assignment to keep compiler quiet */
  FF Fp = NULL; /* assignment to keep compiler quiet */
  FFelem p;
  PrimeSource PS;
  mpq_t *v, *w, *vptr, *wptr;
  FFelem **one_p;
  FFmat **Mmult_p;
  Qmat M, RHS, GBtails;

  one_p = (FFelem**)MALLOC(nideals*sizeof(FFelem*));
  for (i=0; i < nideals; i++)
    one_p[i] = (FFelem*)MALLOC(quotdim[i]*sizeof(FFelem));
  Mmult_p = (FFmat**)MALLOC(nideals*sizeof(FFmat*));
  for (i=0; i < nideals; i++)
  {
    Mmult_p[i] = (FFmat*)MALLOC(nvars*sizeof(FFmat));
    for (x=0; x < nvars; x++)
      Mmult_p[i][x] = FFmat_ctor(quotdim[i], quotdim[i]);
  }
  PS = PrimeSourceCtor();
  try_another_prime:
  for (p = NextPrime(PS); p != 0; p = NextPrime(PS))
  {
    /* Convert one to one_p, Mmult to Mmult_p */
    Fp = FFctor(p);
    FFselect(Fp);
    for (i=0; i < nideals; i++)
    {
      for (j=0; j < quotdim[i]; j++)
      {
        OK = mpq_to_FFelem(&one_p[i][j], one[i][j]);
        if (!OK) goto failed;
      }
    }
    for (i=0; i < nideals; i++)
    {
      for (x=0; x < nvars; x++)
      {
        OK = Qmat_to_FFmat(Mmult_p[i][x], Mmult[i][x]);
        if (!OK) goto failed;
      }
    }
    /* if we get here the prime was good */
    break;
    failed:
    FFselect(NULL);
    FFdtor(Fp);
  }  

  trace = GBM_affine_mod_p(nvars, nideals, quotdim, one_p, Mmult_p, cmp);
  FFdtor(Fp);
  ncols = trace->ncols;
  nseps = trace->nseps;
  pp_col = (int**)MALLOC(nseps*sizeof(int*));
  j = 0;
  for (i=0; i < nseps; i++)
  {
    while (trace->sep[j] < 0) j++;
    pp_col[i] = trace->pp[trace->sep[j]];
    j++;
  }

  GBsize = trace->GBsize;
  v = (mpq_t*)MALLOC(ncols*sizeof(mpq_t));
  for (i=0; i < ncols; i++) mpq_init(v[i]);
  w = (mpq_t*)MALLOC(ncols*sizeof(mpq_t));
  for (i=0; i < ncols; i++) mpq_init(w[i]);

  M = Qmat_ctor(ncols, nseps);
  /* First column is always full of ones */
  for (i=0; i < ncols; i++)
    mpq_set_ui(M->entry[i][0], 1, 1);

  for (j=1; j < nseps; j++)
  {
    /* Find "parent" of current PP, which may appear later in the list! */
    for (t=0; t < nseps; t++)
      if (t != j && (x = pp_parent(nvars, pp_col[t], pp_col[j])) >= 0) break;

    for (i=0; i < ncols; i++) mpq_set(v[i], M->entry[i][t]);
    wptr = w;
    vptr = v;
    for (i=0; i < nideals; i++)
    {
      Qmat_mul_vec(wptr, Mmult[i][x], vptr);
      wptr += quotdim[i];
      vptr += quotdim[i];
    }
    for (k=0; k < ncols; k++)
      mpq_set(M->entry[k][j], w[k]);
  }
  RHS = Qmat_ctor(ncols, GBsize);
  for (j=0; j < GBsize; j++)
  {
    s = trace->GB[j];
    for (t=0; t < nseps; t++)
      if ((x = pp_parent(nvars, pp_col[t], trace->pp[s])) >= 0) break;
    for (i=0; i < ncols; i++) mpq_set(v[i], M->entry[i][t]);
    wptr = w;
    vptr = v;
    for (i=0; i < nideals; i++)
    {
      Qmat_mul_vec(wptr, Mmult[i][x], vptr);
      wptr += quotdim[i];
      vptr += quotdim[i];
    }
    for (k=0; k < ncols; k++)
      mpq_neg(RHS->entry[k][j], w[k]); /* negate to make it easier to build GBasis elems in jintf.c */
  }

  for (i=0; i < nideals; i++)
  {
    for (x=0; x < nvars; x++) FFmat_dtor(Mmult_p[i][x]);
    FREE(Mmult_p[i]);
  }
  FREE(Mmult_p);
  for (i=0; i < nideals; i++) FREE(one_p[i]);
  FREE(one_p);
  for (i=0; i < ncols; i++) mpq_clear(w[i]);
  FREE(w);
  for (i=0; i < ncols; i++) mpq_clear(v[i]);
  FREE(v);
  FREE(pp_col);
  GBtails = Qmat_ctor(nseps, GBsize);
  Qsolve(GBtails, M, RHS);
  /* Must check if we get zeros where we should; if not, go to try_another_prime */
*trace_out = trace;
*GBtails_out = GBtails;
Qmat_dtor(RHS);Qmat_dtor(M);
PrimeSourceDtor(PS);
}



