/* test mostly for distinct degree factorization, but there are some more
   stringent tests for squarefree decomposition too */

#include <stdio.h>

#include "jalloc.h"
#include "DUPFF.h"
#include "DUPFFprint.h"
#include "DUPFFddf.h"
#include "jerror.h"
#include "DUPFFlist.h"
#include "DUPFFsqfrd.h"


void singletest(FFelem p, int d);

int main(int argc, char **argv)
{
  singletest(19, 5);
  singletest(19, 17);
  singletest(19, 18);
  singletest(97, 5);
  singletest(97, 25);
  singletest(7, 19);
  singletest(3, 37);
  singletest(3, 38);

  if (LAST_ERROR != 0)
    printf("********ERROR %d OCCURRED********\n", LAST_ERROR);
  return(0);
}

void check(const DUPFF f)
{
  int *ddfdeg;
  DUPFF tmp, ddfcmpt;
  int i, j;
  DUPFFlist sqfr_cmpts, iter, sqfr_cmpts2;

  sqfr_cmpts = DUPFFsqfrd(f);
  printf("number of sqfr components is %d\n", DUPFFlist_length(sqfr_cmpts));
  printf("Shape is: ");
  for (iter=sqfr_cmpts; iter; iter = iter->next)
  {
    printf("P(deg=%d)^%d ", DUPFFdeg(iter->poly), iter->deg);
  }
  printf("\n");

  /* check that sqfr decomp is a true factorization */
  tmp = DUPFFnew(DUPFFdeg(f));
  tmp->deg = 0;
  tmp->coeffs[0] = 1;
  for (iter=sqfr_cmpts; iter; iter = iter->next)
    for (j=0; j < iter->deg; j++)
      DUPFFmul3(tmp, tmp, iter->poly);
  /* NB product is correct only up to a scalar */
  if (DUPFFdeg(tmp) != DUPFFdeg(f))
    printf("********SQFRD DEGREE ERROR********\n");
  DUPFFrem2(tmp, f);
  if (!DUPFFdeg(tmp) < 0)
    printf("********SQFRD IS WRONG**********\n");

  /* check each sqfr component doesn't decompose */
  for (iter=sqfr_cmpts; iter; iter = iter->next)
  {
    sqfr_cmpts2 = DUPFFsqfrd(iter->poly);
    if (DUPFFlist_length(sqfr_cmpts2) != 1 || sqfr_cmpts2->deg != 1)
      printf("********sqfr component isn't sqfr********\n");
    if (DUPFFdeg(sqfr_cmpts2->poly) != DUPFFdeg(iter->poly))
      printf("********sqfr component deg changed********\n");
    DUPFFrem2(sqfr_cmpts2->poly, iter->poly);
    if (DUPFFdeg(sqfr_cmpts2->poly) >= 0)
      printf("********sqfr component doesn't divide itself********\n");
    DUPFFlist_dtor(sqfr_cmpts2);
  }

  /* now compute and check DDF for each sqfr component */
  for (i=0, iter=sqfr_cmpts; iter; iter = iter->next, i++)
  {
    printf("DDF for sqfr component %d: ", i);
    DUPFFddf(&ddfcmpt, &ddfdeg, iter->poly);
    tmp->deg = 0;
    tmp->coeffs[0] = 1;
    for (j=0; ddfdeg[j] != 0; j++)
    {
      if ((DUPFFdeg(ddfcmpt+j) % ddfdeg[j]) != 0)
	printf("********DEG ERROR IN DDF********\n");
      if (ddfdeg[j] == DUPFFdeg(ddfcmpt+j))
	printf("Irr(%d) ", ddfdeg[j]);
      else
	printf("Cmp(%d/%d) ", DUPFFdeg(ddfcmpt+j), ddfdeg[j]);
      DUPFFmul3(tmp, tmp, ddfcmpt+j);
    }
    printf("\n");
    if (DUPFFdeg(tmp) != DUPFFdeg(iter->poly))
      printf("********DEG ERROR IN DDF********\n");
    DUPFFrem2(tmp, iter->poly);
    if (!DUPFFdeg(tmp) < 0)
      printf("********DDF IS WRONG**********\n");

    /* clean up allocated memory */
    for (j=0; ddfdeg[j] != 0; j++) FREE((ddfcmpt+j)->coeffs);
    FREE(ddfcmpt);
    FREE(ddfdeg);
  }

  /* clean up allocated memory */
  DUPFFlist_dtor(sqfr_cmpts);
  DUPFFfree(tmp);

  printf("Check complete.\n\n\n");
}


void singletest(FFelem p, int d)
{
  FF K;
  DUPFF f, g, h, tmp;
  int i;

  printf("Starting test for p=%d and d=%d\n", p, d);
  K = FFctor(p);
  FFselect(K);
  f = DUPFFnew(d);
  g = DUPFFnew(d);
  h = DUPFFnew(d);
  for (i=0; i<=d; i++) f->coeffs[i] = (4*i*i+7*i+3)%p;
  for (i=0; i<=d; i++) g->coeffs[i] = (i*i+17*i+9)%p;
  for (i=0; i<=d; i++) h->coeffs[i] = (i*i*i+3*i*i+7*i+4)%p;
  f->deg = d; if (f->coeffs[d] == 0) f->coeffs[d] = p/3+1;
  g->deg = d; if (g->coeffs[d] == 0) g->coeffs[d] = 2*p/3-1;
  h->deg = d; if (h->coeffs[d] == 0) h->coeffs[d] = 3*p/5+1;

  printf("f is ");
  DUPFFprint(f);
  printf("g is ");
  DUPFFprint(g);
  printf("h is ");
  DUPFFprint(h);

  printf("checking sqfrd for f...\n");
  check(f);
  printf("checking sqfrd for g...\n");
  check(g);
  printf("checking sqfrd for h...\n");
  check(h);

  tmp = DUPFFnew(6*d);
  DUPFFcopy2(tmp, f);
  DUPFFmul3(tmp, tmp, g);
  DUPFFmul3(tmp, tmp, g);
  DUPFFmul3(tmp, tmp, h);
  DUPFFmul3(tmp, tmp, h);
  DUPFFmul3(tmp, tmp, h);
  printf("checking sqfrd for f*g^2*h^3...\n");
  check(tmp);
  DUPFFfree(tmp);

  DUPFFfree(f);
  DUPFFfree(g);
  DUPFFfree(h);
  FFdtor(K);
  
  printf("test for p=%d and d=%d complete\n\n\n\n\n\n", p, d);
}
